@@ -33,6 +33,7 @@
 #include "flatpak-builtins-utils.h"
 #include "flatpak-utils-private.h"
 #include "flatpak-table-printer.h"
+#include "flatpak-variant-impl-private.h"
 
 static char *opt_arch;
 static char *opt_commit;
@@ -46,6 +47,7 @@ static gboolean opt_log;
 static gboolean opt_show_runtime;
 static gboolean opt_show_sdk;
 static gboolean opt_cached;
+static gboolean opt_sideloaded;
 
 static GOptionEntry options[] = {
   { "arch", 0, 0, G_OPTION_ARG_STRING, &opt_arch, N_("Arch to install for"), N_("ARCH") },
@@ -60,6 +62,8 @@ static GOptionEntry options[] = {
   { "show-runtime", 0, 0, G_OPTION_ARG_NONE, &opt_show_runtime, N_("Show runtime"), NULL },
   { "show-sdk", 0, 0, G_OPTION_ARG_NONE, &opt_show_sdk, N_("Show sdk"), NULL },
   { "cached", 0, 0, G_OPTION_ARG_NONE, &opt_cached, N_("Use local caches even if they are stale"), NULL },
+  /* Translators: A sideload is when you install from a local USB drive rather than the Internet. */
+  { "sideloaded", 0, 0, G_OPTION_ARG_NONE, &opt_sideloaded, N_("Only list refs available as sideloads"), NULL },
   { NULL }
 };
 
@@ -79,36 +83,32 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
   g_autoptr(GPtrArray) dirs = NULL;
   g_autoptr(FlatpakDir) preferred_dir = NULL;
   g_autoptr(GVariant) commit_v = NULL;
-  g_autoptr(GVariant) commit_metadata = NULL;
   const char *remote;
   const char *pref;
   g_autofree char *default_branch = NULL;
   FlatpakKinds kinds;
   FlatpakKinds matched_kinds;
-  g_autofree char *id = NULL;
-  g_autofree char *arch = NULL;
-  g_autofree char *branch = NULL;
-  g_auto(GStrv) parts = NULL;
-  FlatpakKinds kind;
-  g_autofree char *ref = NULL;
+  g_autofree char *match_id = NULL;
+  g_autofree char *match_arch = NULL;
+  g_autofree char *match_branch = NULL;
+  g_autoptr(FlatpakDecomposed) ref = NULL;
   g_autofree char *commit = NULL;
   g_autofree char *parent = NULL;
   g_autoptr(FlatpakRemoteState) state = NULL;
-  g_autoptr(GVariant) sparse = NULL;
-  const char *eol = NULL;
-  const char *eol_rebase = NULL;
   gboolean friendly = TRUE;
   const char *xa_metadata = NULL;
   const char *collection_id = NULL;
+  const char *eol = NULL;
+  const char *eol_rebase = NULL;
   g_autoptr(GKeyFile) metakey = NULL;
   guint64 installed_size = 0;
   guint64 download_size = 0;
   g_autofree char *formatted_installed_size = NULL;
   g_autofree char *formatted_download_size = NULL;
   const gchar *subject = NULL;
-  const gchar *body = NULL;
   guint64 timestamp;
   g_autofree char *formatted_timestamp = NULL;
+  VarMetadataRef sparse_cache;
 
   context = g_option_context_new (_(" REMOTE REF - Show information about an application or runtime in a remote"));
   g_option_context_set_translation_domain (context, GETTEXT_PACKAGE);
@@ -133,54 +133,53 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
   kinds = flatpak_kinds_from_bools (opt_app, opt_runtime);
 
   if (!flatpak_split_partial_ref_arg (pref, kinds, opt_arch, NULL,
-                                      &matched_kinds, &id, &arch, &branch, error))
+                                      &matched_kinds, &match_id, &match_arch, &match_branch, error))
     return FALSE;
 
-  ref = flatpak_dir_find_remote_ref (preferred_dir, remote, id, branch, default_branch, arch,
-                                     matched_kinds, &kind, cancellable, error);
-  if (ref == NULL)
-    return FALSE;
-
-  state = get_remote_state (preferred_dir, remote, opt_cached, cancellable, error);
+  state = get_remote_state (preferred_dir, remote, opt_cached, opt_sideloaded, match_arch, NULL, NULL, error);
   if (state == NULL)
     return FALSE;
 
+  ref = flatpak_dir_find_remote_ref (preferred_dir, state, match_id, match_branch, default_branch, match_arch,
+                                     matched_kinds, cancellable, error);
+  if (ref == NULL)
+    return FALSE;
+
   if (opt_cached)
     {
       if (opt_commit)
         commit = g_strdup (opt_commit);
       else
         {
-          flatpak_remote_state_lookup_ref (state, ref, &commit, NULL, error);
+          flatpak_remote_state_lookup_ref (state, flatpak_decomposed_get_ref (ref), &commit, NULL, NULL, NULL, error);
           if (commit == NULL)
             {
               if (error != NULL && *error == NULL)
                 flatpak_fail_error (error, FLATPAK_ERROR_REF_NOT_FOUND,
                                     _("Couldn't find latest checksum for ref %s in remote %s"),
-                                    ref, remote);
+                                    flatpak_decomposed_get_ref (ref), remote);
               return FALSE;
             }
         }
     }
   else
     {
-      commit_v = flatpak_dir_fetch_remote_commit (preferred_dir, remote, ref, opt_commit, &commit, cancellable, error);
+      commit_v = flatpak_remote_state_load_ref_commit (state, preferred_dir, flatpak_decomposed_get_ref (ref),
+                                                       opt_commit, NULL, &commit, cancellable, error);
       if (commit_v == NULL)
         return FALSE;
     }
 
-  sparse = flatpak_remote_state_lookup_sparse_cache (state, ref, NULL);
-  if (sparse)
+  if (flatpak_remote_state_lookup_sparse_cache (state, flatpak_decomposed_get_ref (ref),
+                                                &sparse_cache, NULL))
     {
-      g_variant_lookup (sparse, FLATPAK_SPARSE_CACHE_KEY_ENDOFLINE, "&s", &eol);
-      g_variant_lookup (sparse, FLATPAK_SPARSE_CACHE_KEY_ENDOFLINE_REBASE, "&s", &eol_rebase);
+      eol = var_metadata_lookup_string (sparse_cache, FLATPAK_SPARSE_CACHE_KEY_ENDOFLINE, NULL);
+      eol_rebase = var_metadata_lookup_string (sparse_cache, FLATPAK_SPARSE_CACHE_KEY_ENDOFLINE_REBASE, NULL);
     }
 
   if (opt_show_ref || opt_show_commit || opt_show_parent || opt_show_metadata || opt_show_runtime || opt_show_sdk)
     friendly = FALSE;
 
-  parts = g_strsplit (ref, "/", 0);
-
   if (friendly)
     {
       int len;
@@ -190,6 +189,7 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
       AsApp *app = NULL;
       const char *version = NULL;
       const char *license = NULL;
+      g_autofree char *id = flatpak_decomposed_dup_id (ref);
 
       flatpak_get_window_size (&rows, &cols);
 
@@ -197,8 +197,8 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
       as_store_set_add_flags (store, as_store_get_add_flags (store) | AS_STORE_ADD_FLAG_USE_UNIQUE_ID);
 #endif
 
-      flatpak_dir_load_appstream_store (preferred_dir, remote, parts[2], store, NULL, NULL);
-      app = as_store_find_app (store, ref);
+      flatpak_dir_load_appstream_store (preferred_dir, remote, id, store, NULL, NULL);
+      app = as_store_find_app (store, flatpak_decomposed_get_ref (ref));
       if (app)
         {
           const char *name = as_app_get_localized_name (app);
@@ -212,14 +212,16 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
 
       if (commit_v)
         {
-          g_variant_get (commit_v, "(a{sv}aya(say)&s&stayay)", NULL, NULL, NULL,
-                         &subject, &body, NULL, NULL, NULL);
+          VarCommitRef var_commit = var_commit_from_gvariant (commit_v);
+          VarMetadataRef commit_metadata;
 
+          subject = var_commit_get_subject (var_commit);
           parent = ostree_commit_get_parent (commit_v);
           timestamp = ostree_commit_get_timestamp (commit_v);
 
-          commit_metadata = g_variant_get_child_value (commit_v, 0);
-          g_variant_lookup (commit_metadata, "xa.metadata", "&s", &xa_metadata);
+          commit_metadata = var_commit_get_metadata (var_commit);
+          xa_metadata = var_metadata_lookup_string (commit_metadata, "xa.metadata", NULL);
+
           if (xa_metadata == NULL)
             g_printerr (_("Warning: Commit has no flatpak metadata\n"));
           else
@@ -229,13 +231,10 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
                 return FALSE;
             }
 
-          g_variant_lookup (commit_metadata, "ostree.collection-binding", "&s", &collection_id);
+          collection_id = var_metadata_lookup_string (commit_metadata, "ostree.collection-binding", NULL);
 
-          if (g_variant_lookup (commit_metadata, "xa.installed-size", "t", &installed_size))
-            installed_size = GUINT64_FROM_BE (installed_size);
-
-          if (g_variant_lookup (commit_metadata, "xa.download-size", "t", &download_size))
-            download_size = GUINT64_FROM_BE (download_size);
+          installed_size = GUINT64_FROM_BE (var_metadata_lookup_uint64 (commit_metadata, "xa.installed-size", 0));
+          download_size = GUINT64_FROM_BE (var_metadata_lookup_uint64 (commit_metadata, "xa.download-size", 0));
 
           formatted_installed_size = g_format_size (installed_size);
           formatted_download_size = g_format_size (download_size);
@@ -257,7 +256,7 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
         len = MAX (len, g_utf8_strlen (_("Download:"), -1));
       if (formatted_installed_size)
         len = MAX (len, g_utf8_strlen (_("Installed:"), -1));
-      if (strcmp (parts[0], "app") == 0 && metakey != NULL)
+      if (flatpak_decomposed_is_app (ref) == 0 && metakey != NULL)
         {
           len = MAX (len, g_utf8_strlen (_("Runtime:"), -1));
           len = MAX (len, g_utf8_strlen (_("Sdk:"), -1));
@@ -278,10 +277,10 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
 
       width = cols - (len + 1);
 
-      print_aligned (len, _("ID:"), parts[1]);
-      print_aligned (len, _("Ref:"), ref);
-      print_aligned (len, _("Arch:"), parts[2]);
-      print_aligned (len, _("Branch:"), parts[3]);
+      print_aligned (len, _("ID:"), id);
+      print_aligned (len, _("Ref:"), flatpak_decomposed_get_ref (ref));
+      print_aligned_take (len, _("Arch:"), flatpak_decomposed_dup_arch (ref));
+      print_aligned (len, _("Branch:"), flatpak_decomposed_dup_arch (ref));
       if (version != NULL)
         print_aligned (len, _("Version:"), version);
       if (license != NULL)
@@ -292,17 +291,17 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
         print_aligned (len, _("Download:"), formatted_download_size);
       if (formatted_installed_size)
         print_aligned (len, _("Installed:"), formatted_installed_size);
-      if (strcmp (parts[0], "app") == 0 && metakey != NULL)
+      if (flatpak_decomposed_is_app (ref) && metakey != NULL)
         {
           g_autofree char *runtime = g_key_file_get_string (metakey, "Application", "runtime", error);
           print_aligned (len, _("Runtime:"), runtime ? runtime : "-");
         }
-      g_print ("\n");
-      if (strcmp (parts[0], "app") == 0 && metakey != NULL)
+      if (flatpak_decomposed_is_app (ref) && metakey != NULL)
         {
           g_autofree char *sdk = g_key_file_get_string (metakey, "Application", "sdk", error);
           print_aligned (len, _("Sdk:"), sdk ? sdk : "-");
         }
+      g_print ("\n");
       {
         g_autofree char *formatted_commit = ellipsize_string (commit, width);
         print_aligned (len, _("Commit:"), formatted_commit);
@@ -341,8 +340,10 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
               guint64 p_timestamp;
               g_autofree char *p_formatted_timestamp = NULL;
               g_autoptr(GVariant) p_commit_v = NULL;
+              VarCommitRef p_commit;
 
-              p_commit_v = flatpak_dir_fetch_remote_commit (preferred_dir, remote, ref, p, NULL, cancellable, NULL);
+              p_commit_v = flatpak_remote_state_load_ref_commit (state, preferred_dir, flatpak_decomposed_get_ref (ref),
+                                                                 p, NULL, NULL, cancellable, NULL);
               if (p_commit_v == NULL)
                 break;
 
@@ -350,8 +351,8 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
               p_timestamp = ostree_commit_get_timestamp (p_commit_v);
               p_formatted_timestamp = format_timestamp (p_timestamp);
 
-              g_variant_get (p_commit_v, "(a{sv}aya(say)&s&stayay)", NULL, NULL, NULL,
-                             &p_subject, NULL, NULL, NULL, NULL);
+              p_commit = var_commit_from_gvariant (p_commit_v);
+              p_subject = var_commit_get_subject (p_commit);
 
               print_aligned (len, _(" Commit:"), p);
               print_aligned (len, _(" Subject:"), p_subject);
@@ -400,7 +401,7 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
           if (opt_show_ref)
             {
               maybe_print_space (&first);
-              g_print ("%s", ref);
+              g_print ("%s", flatpak_decomposed_get_ref (ref));
             }
 
           if (opt_show_commit)
@@ -421,12 +422,7 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
               maybe_print_space (&first);
 
               if (metakey)
-                {
-                  if (strcmp (parts[0], "app") == 0)
-                    runtime = g_key_file_get_string (metakey, "Application", "runtime", NULL);
-                  else
-                    runtime = g_key_file_get_string (metakey, "Runtime", "runtime", NULL);
-                }
+                runtime = g_key_file_get_string (metakey, flatpak_decomposed_get_kind_metadata_group (ref), "runtime", NULL);
               g_print ("%s", runtime ? runtime : "-");
             }
 
@@ -436,12 +432,7 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
               maybe_print_space (&first);
 
               if (metakey)
-                {
-                  if (strcmp (parts[0], "app") == 0)
-                    sdk = g_key_file_get_string (metakey, "Application", "sdk", NULL);
-                  else
-                    sdk = g_key_file_get_string (metakey, "Runtime", "sdk", NULL);
-                }
+                sdk = g_key_file_get_string (metakey, flatpak_decomposed_get_kind_metadata_group (ref), "sdk", NULL);
               g_print ("%s", sdk ? sdk : "-");
             }
 
@@ -463,7 +454,9 @@ flatpak_builtin_remote_info (int argc, char **argv, GCancellable *cancellable, G
           c_v = NULL;
 
           if (c && opt_log)
-            c_v = flatpak_dir_fetch_remote_commit (preferred_dir, remote, ref, c, NULL, cancellable, NULL);
+            c_v = flatpak_remote_state_load_ref_commit (state, preferred_dir,
+                                                        flatpak_decomposed_get_ref (ref),
+                                                        c, NULL, NULL, cancellable, NULL);
         }
       while (c_v != NULL);
     }
